using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

using GameMenu;

namespace GameOfLife
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;

        static int WIDTH = 64;
        static int HEIGHT = 48;
        Rectangle WINDOWRECT;

        enum GameSpeed
        {
            Slow,
            Medium,
            Fast
        };
        GameSpeed m_speed;

        // if this is set, the game will run continuously
        Boolean runFreely;

        Menu menu;
        static string MENU_CLEAR_FIELD = "Clear Field";
        static string MENU_GAME_SPEED = "Game Speed -->";
        static string STARTING_CONDITIONS = "Starting Conditions";
        static string RETURN_TO_GAME = "Return";

        // initialized in loadContent
        int POINTHEIGHT;
        int POINTWIDTH;

        LifeField m_lifeField;
        Texture2D m_particleTexture;
        Texture2D m_mouseTexture;

        Vector2 m_mousePos;

        double m_lastUpdate;

        #region helper functions
        private long GetUpdateIntervalMs()
        {
            switch (m_speed)
            {
                case GameSpeed.Slow:
                    return 100;
                case GameSpeed.Medium:
                    return 50;
                case GameSpeed.Fast:
                    return 25;
            }
            return 50;
        }
     
        private Boolean NeedToUpdateField(GameTime time)
        {
            if ((time.TotalRealTime.TotalMilliseconds - m_lastUpdate) > GetUpdateIntervalMs())
            {
                return true;
            }

            return false;
        }
        #endregion

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            graphics.PreferredBackBufferHeight = 768;
            graphics.PreferredBackBufferWidth = 1024;
            WINDOWRECT = new Rectangle(0, 0, 1024, 768);
            graphics.PreferMultiSampling = false;
            Content.RootDirectory = "Content";
            m_speed = GameSpeed.Medium;
            m_lastUpdate = 0;
            runFreely = false;
         }

        public void ChoiceExecuted(object source, Menu.MenuEvent e)
        {
            if (e.choiceString == MENU_CLEAR_FIELD)
            {
                m_lifeField.ClearField();
                runFreely = false;
                menu.visible = false;
            }
            else if (e.choiceString == RETURN_TO_GAME)
            {
                menu.visible = false;
            }
            Console.WriteLine("executed {0}", e.choiceString);
        }

        public void ChoiceSelected(object source, Menu.MenuEvent e)
        {
            if (e.choiceString == "Slow")
                m_speed = GameSpeed.Slow;
            else if (e.choiceString == "Medium")
                m_speed = GameSpeed.Medium;
            else if (e.choiceString == "Fast")
                m_speed = GameSpeed.Fast;
            
            Console.WriteLine("selected {0}", e.choiceString);
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            m_lifeField = new LifeField(WIDTH, HEIGHT);
            spriteBatch = new SpriteBatch(graphics.GraphicsDevice);

            m_mousePos = Vector2.Zero;

            menu = new Menu(this);
            menu.AddChoice(MENU_CLEAR_FIELD);
            menu.AddChoice(MENU_GAME_SPEED).AddLeftRightChoices(new string[] { "Slow", "Medium", "Fast" });

            // code i decided not to write.... yet
            //MenuChoice conditions = menu.AddChoice(STARTING_CONDITIONS);
            // conditions.AddChoice("Gun");
            // conditions.AddChoice("Long Haul");

            menu.ChoiceExecuted += new Menu.ChoiceExecutedHandler(ChoiceExecuted);
            menu.ChoiceSelected += new Menu.ChoiceSelectedHandler(ChoiceSelected);

            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            POINTHEIGHT = graphics.PreferredBackBufferHeight / HEIGHT;
            POINTWIDTH = graphics.PreferredBackBufferWidth / WIDTH;

            m_particleTexture = new Texture2D(graphics.GraphicsDevice, (int)POINTWIDTH, (int)POINTHEIGHT, 1, TextureUsage.None, SurfaceFormat.Color);

            Color[] texture = new Color[POINTWIDTH * POINTHEIGHT];
            for (int i = 0; i < texture.Length; ++i)
            {
                texture[i] = Color.Red;
            }
            m_particleTexture.SetData(texture);

            m_mouseTexture = new Texture2D(graphics.GraphicsDevice, (int)POINTWIDTH, (int)POINTHEIGHT, 1, TextureUsage.None, SurfaceFormat.Color);

            texture = new Color[POINTWIDTH * POINTHEIGHT];
            for (int i = 0; i < texture.Length; ++i)
            {
                texture[i] = Color.Blue;
            }
            m_mouseTexture.SetData(texture);

            // do the first update so we
            m_lifeField.Update(new GameTime());
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        KeyboardState prevKbState;
        MouseState prevMouseState;

        // value we're toggling cells to.  This is so we can press and hold the mouse
        // and we will be setting every cell it passes over to one value.  This keeps
        // us from toggling a cell each time we cross over it with the mouse.
        Boolean currentSetToValue; 

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            // Allows the game to exit
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
                this.Exit();

            KeyboardState kbState = Keyboard.GetState();
            MouseState mouseState = Mouse.GetState();

            m_mousePos.X = mouseState.X;
            m_mousePos.Y = mouseState.Y;

            if (kbState.IsKeyDown(Keys.Escape) && !prevKbState.IsKeyDown(Keys.Escape))
            {
                // show the menu
                menu.visible = !menu.visible;
            }
            // allow no action here until menu is hidden
            if (menu.visible)
            {
                prevKbState = kbState;
                prevMouseState = mouseState;
                menu.Update(gameTime);
                return;
            }

            if (kbState.IsKeyDown(Keys.Enter) && !prevKbState.IsKeyDown(Keys.Enter))
            {
                // let the game run
                runFreely = !runFreely;
            }

            if (mouseState.LeftButton == ButtonState.Pressed)
            {
                int mouseX = (int)m_mousePos.X / POINTWIDTH;
                int mouseY = (int)m_mousePos.Y / POINTHEIGHT;
                Boolean value = m_lifeField[mouseX, mouseY];
                
                // if they released the mouse, then reset the state we're setting cells to.
                if(prevMouseState.LeftButton == ButtonState.Released)
                    currentSetToValue = !value;

                m_lifeField[mouseX, mouseY] = currentSetToValue;
            }

            if (kbState.IsKeyDown(Keys.Space) || runFreely)
            {
                if (NeedToUpdateField(gameTime))
                {
                    m_lifeField.Update(gameTime);
                    m_lastUpdate = gameTime.TotalRealTime.TotalMilliseconds;
                }
            }

            prevKbState = kbState;
            prevMouseState = mouseState;
            base.Update(gameTime);
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            graphics.GraphicsDevice.Clear(Color.CornflowerBlue);

            Texture2D colorMap = new Texture2D(graphics.GraphicsDevice, WIDTH, HEIGHT, 1, 
                            TextureUsage.Linear, SurfaceFormat.Color);
            colorMap.SetData(m_lifeField.GetColorMap());
            
            spriteBatch.Begin(SpriteBlendMode.AlphaBlend, SpriteSortMode.Immediate, SaveStateMode.None);
            graphics.GraphicsDevice.SamplerStates[0].MagFilter = TextureFilter.None;

            spriteBatch.Draw(colorMap, WINDOWRECT, Color.White);

            
            spriteBatch.End();

            menu.Draw(gameTime);

            spriteBatch.Begin();
            // draw the mouse on top of everything
            spriteBatch.Draw(m_mouseTexture, m_mousePos, null, Color.White, 0f,
                    Vector2.Zero, 1.0f, SpriteEffects.None, 0f);
            spriteBatch.End();
            

            base.Draw(gameTime);
        }
    }
}
